/***************************************************************************
 *   Copyright (C) 2010 by Ahsan Tariq   *
 *   ahsan@mpi-sws.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <vector>
#include <string>
#include <cstring>
#include <cstdlib>
#include <fstream>

#include "RIB.h"
#include "Algorithms.h"

extern bool ALL_UPDATES_POTENT;

using namespace std;

/**********************************************************************************************************************
Constructor for RIB. Creates the head of the RIB (root).
initializes number NodesChanged to zero
*********************************************************************************************************************/

cRIB::cRIB(cPolicy *p, int s)	{
			pol = p;
			root = new cRIBNode (0,0,NULL,NULL,NULL,pol);
			SMALTAEnabled = 0;
			ZhangAggEnabled = 0;
			NodesChanged = 0;
		}

/*************************************************************************************************************************
Constructor for class cRIBNode	

Paramters:
INFIB: Fills the INFIB flag
Original: Fills the original flag
par: Pointer to parent.
lchild: Pointer to left child
rchild: Pointer to right child
p: policy of selecting best nexthops
************************************************************************************************************************/

cRIBNode::cRIBNode (int INFIB, int Original, cRIBNode *par, cRIBNode *lchild, cRIBNode *rchild, cPolicy *p)	{

	Data.RIBFlags.INFIB = INFIB;
	parent = par;
	child[LEFT] = lchild;
	child[RIGHT] = rchild;
	Data.Routes.BestNH = NULL_NEXTHOP;
	Data.Routes.BestASLen = 0;
	pol = p;
	Data.SMALTAFlags.Original = Original;
}



/************************************************************************************************************************
SetRoutes function selects the Best NextHop or adds the NextHop to the Nexthop vector. Best NextHop depends on user
policy. Also checks if the nexthop already exists so that no duplicates are created.

Returns:
0	if Best Nexthop remains unchanged
1	is never returned
2	indicates a new best nexthop selected from alternate routes
3	indicates the nexthop announced was considered the best in the abscence of alternate routes
	NOTE: Contrast with 2 in which alternate routes were present. 
	NOTE: Differentiation between 3 and 2 is needed for the SMALTA update algorithm
4	Indicates a non-original and INFIB entry was present. The entry was removed and replaced by the announced nexthop

Parameters:

NH: The announced nexthop
ASLen: The corresponding AS path length
SMALTAEnabled: whether SMALTA is enabled or not. Behaviour changes if it is enabled
ZhangAggEnabled: whether Level (1 - 4) is enabled or not. Behaviour changes if it is enabled.
************************************************************************************************************************/


int cRIBNode::SetRoutes (unsigned int NH, int ASLen, int SMALTAEnabled, int ZhangAggEnabled)	
{
  if(ALL_UPDATES_POTENT)
    {
      // the following two lines should be executed for ATT only
      Data.Routes.BestNH = NULL_NEXTHOP;
      Data.Routes.BestASLen = 1000; // something infinite!
    }

  unsigned int CurrBestNH = Data.Routes.BestNH;
  int CurrBestASLen = Data.Routes.BestASLen;
  int index = -1;
	
  if (SMALTAEnabled == 1 || ZhangAggEnabled == 1)	
    {
      if (Data.SMALTAFlags.Original == 0 && Data.Routes.BestNH != NULL_NEXTHOP)	
	{
	  Data.SMALTAFlags.Original = 1;
	  Data.SMALTAFlags.DistanceToParent = -1;
	  Data.RIBFlags.INFIB = 0;
	  Data.Routes.BestNH = NH;
	  Data.Routes.BestASLen = ASLen;
	  return 4;
	}
    }

  if (CurrBestNH == NULL_NEXTHOP)
    {
      Data.Routes.BestNH = NH;
      Data.Routes.BestASLen = ASLen;
      return 3; // announce to the FIB
    }
  else if (CurrBestNH == NH)	
    {
      if (CurrBestASLen == ASLen)
	return 0; // nothing changed in our best
      else if (CurrBestASLen > ASLen)	
	{
	  // our existing best just got improved
	  Data.Routes.BestASLen = ASLen;
	  return 0;
	}
      else if (CurrBestASLen < ASLen)	
	{
	  // length of the existing best has increased
	  // need to check if some other route has become better
	  CurrBestASLen = ASLen;
	  for (unsigned int i = 0; i < Data.Routes.NHASLenPair.size(); i++)	
	    {
	      if (CurrBestASLen > Data.Routes.NHASLenPair[i].ASLen)	
		{
		  index = i;
		  CurrBestASLen = Data.Routes.NHASLenPair[i].ASLen;
		}
	    }
	  if (index != -1)	
	    {
	      Data.Routes.BestNH = Data.Routes.NHASLenPair[index].NH;
	      Data.Routes.BestASLen = Data.Routes.NHASLenPair[index].ASLen;
	      Data.Routes.NHASLenPair.erase (Data.Routes.NHASLenPair.begin() + index);
	      sNHASLenPair pair;
	      pair.NH = NH;
	      pair.ASLen = ASLen;
	      Data.Routes.NHASLenPair.push_back (pair);
	      return 2; // a change in best NH. Announce to the FIB
	    }
	  else	
	    {
	      Data.Routes.BestASLen = ASLen;
	      return 0;
	    }
	}
    }
  else // the announced nexthop is differnet form the non-null current best
    {
      for (unsigned int i = 0; i < Data.Routes.NHASLenPair.size(); i++)
	{	// if a change in ASLen of already existing NH
	  if (Data.Routes.NHASLenPair[i].NH == NH)
	    {
	      if (Data.Routes.NHASLenPair[i].ASLen == ASLen)
		return 0;
	      else	
		{ 
		  Data.Routes.NHASLenPair[i].ASLen = ASLen;
		  index = i;
		  break;
		}
	    }
	}
      
      if (index != -1) // the announced route was already in the list of routes
	{
	  if (Data.Routes.NHASLenPair[index].ASLen < CurrBestASLen)	
	    {
	      sNHASLenPair pair;
	      pair.NH = CurrBestNH;
	      pair.ASLen = CurrBestASLen;
	      Data.Routes.NHASLenPair.push_back(pair);
	      Data.Routes.BestNH = Data.Routes.NHASLenPair[index].NH;
	      Data.Routes.BestASLen = Data.Routes.NHASLenPair[index].ASLen;
	      Data.Routes.NHASLenPair.erase(Data.Routes.NHASLenPair.begin() + index);
	      return 2; // a change in best NH. Announce to the FIB
	    }
	}
      else	
	{
	  if (ASLen < CurrBestASLen)	
	    {
	      sNHASLenPair pair;
	      pair.NH = CurrBestNH;
	      pair.ASLen = CurrBestASLen;
	      Data.Routes.NHASLenPair.push_back(pair);
	      Data.Routes.BestNH = NH;
	      Data.Routes.BestASLen = ASLen;
	      return 2; // a change in best NH. Announce to the FIB
	    }
	  else	
	    {
	      sNHASLenPair pair;
	      pair.NH = NH;
	      pair.ASLen = ASLen;
	      Data.Routes.NHASLenPair.push_back(pair);
	      return 0;
	    }
	}
    }
  
  return 0; // no return value expected here
}

/***************************************************************************************************************
The function removes a route corresponding to the peerIP as its nexthop. 
Return Values:
0	if the withdraw was for a node with non-original nexthop, or there was no change in the Best nexthop
1	if the withdraw was for a leaf node and SAMLTA and ZhangAgg are disabled. the return value tells the 
	calling function to delete this node
2	If after the withdraw there were no other alternate routes left
3	if after the withdraw a new best nexthop was selected from alternate routes

Parameters:

PeerIP:	Indicates the nexthop of the route to be withdrawn
SMALTAEnabled: Indicates if SMALTA is enabled. The behaviour of the fuction changes to account for this.
ZhangAggEnabled: Indicates if Level (1 - 4) aggregation algorithms is enbaled and adjusts accordingly
***************************************************************************************************************/ 

int cRIBNode::UnSetRoutes (unsigned int peerIP, int SMALTAEnabled, int ZhangAggEnabled)	
{	
  if (Data.Routes.BestNH != NULL_NEXTHOP)	
    {
      if (SMALTAEnabled == 1 || ZhangAggEnabled == 1)	
	if (Data.SMALTAFlags.Original == 0)
	  return 0;
      
      if(ALL_UPDATES_POTENT || (peerIP == Data.Routes.BestNH))
      	{
	  // the withdraw is for the current best guy
	  //cout << "change in best NH\n";
	  if (Data.Routes.NHASLenPair.size() == 0)	
	    {	
	      Data.Routes.BestNH = NULL_NEXTHOP;
	      Data.Routes.BestASLen = 0;
	      if (child [LEFT] == NULL && child [RIGHT] == NULL)	
		{		
		  if (SMALTAEnabled == 0 && ZhangAggEnabled == 0)	
		    {	// delete the node if Aggregation is disabled
		      if (parent->child[LEFT] == this)
			parent->child[LEFT] = NULL;
		      else if (parent->child[RIGHT] == this)
			parent->child[RIGHT] = NULL;
		      else	
			{
			  cout << "error, cannot delete leaf after withdraw\n";
			  exit(1);
			}
		      return 1;		// message to delete the node
		    }
		  return 2;
		}
	      else	
		{
		  if (SMALTAEnabled == 0 && ZhangAggEnabled == 0)	
		    {
		      // The route is withdrawn and the aggregation is disabled
		      Data.RIBFlags.INFIB = 0;
		      Data.SMALTAFlags.Original = 0;
		    }
		  return 2;		// This is a Withdraw to the FIB.	
		}
	    }
	  else	
	    {			
	      // select a new best nexthop and remove it from the vector.
	      int temp = 0;
	      for (unsigned int i = 0; i < Data.Routes.NHASLenPair.size(); i++)	
		{
		  if (Data.Routes.NHASLenPair[i].ASLen < Data.Routes.NHASLenPair[temp].ASLen)
		    temp = i;
		}
	      Data.Routes.BestNH = Data.Routes.NHASLenPair[temp].NH;
	      Data.Routes.BestASLen = Data.Routes.NHASLenPair[temp].ASLen;
	      Data.Routes.NHASLenPair.erase(Data.Routes.NHASLenPair.begin() + temp);
	      return 3;		// we have selected a new best NH. This is an Announce to the FIB
	    }
	}
      else	
	{			// the best NH remains unchanged
	  for (unsigned int i = 0; i < Data.Routes.NHASLenPair.size(); i++)	
	    {
	      if (Data.Routes.NHASLenPair[i].NH == peerIP)	
		{
		  Data.Routes.NHASLenPair.erase(Data.Routes.NHASLenPair.begin() + i);		  
		  break;
		}
	    }
	  return 0;
	}		
    }
  else 
    return 0;
}

/************************************************************************************************************************
RIBAnnounce will create a RIBNode. Add the Nexthop to the Routes list and get the new best Nexthop.
It will then compare this new best Nexthop with the existing one and will determine if there is any change.

returns 0 always

Parameters:	
Prefix: The 32 bit prefix. 
mask: Length of the prefix
nexthop: nexthop of the prefix
ASLength: AS path length of the prefix

May call SMALTA or Zhang's update algorithms if enabled
************************************************************************************************************************/

int cRIB::RIBAnnounce (unsigned int Prefix, int mask, unsigned int nexthop, int ASLength)	{
        //cout << "Announce\n";
	cRIBNode *next = root;
	int nbit = 32;
	int ret;
	int tmask = mask;
	unsigned int OldNH;
	//if (e > 5495611)
		//cout << Prefix << "\n";
	unsigned int bit = 0;
	//std::cout << Prefix << " " << mask << " " << nexthop << " " << ASLength << "\n";
	
	while (mask > 0)	{
		nbit--;
		bit = (Prefix >> nbit) % 2;
		//std::cout << bit << "\n";
		if ((*next).child[bit] == NULL)	{ 
			cRIBNode* NewNode = new cRIBNode (0, 0, next, NULL, NULL,pol);
			

			(*next).child[bit] = NewNode;
			next = NewNode;
		}
		else	{
			next = (*next).child[bit];
		}
		mask--;
	}
	OldNH = (*next).Data.Routes.BestNH;
	if (SMALTAEnabled == 0 && ZhangAggEnabled == 0)	{
		next->Data.RIBFlags.INFIB = 1;
		next->Data.SMALTAFlags.Original = 1;
	}
	ret = (*next).SetRoutes (nexthop, ASLength, SMALTAEnabled, ZhangAggEnabled);
 	if (ret == 2)	{
		if (SMALTAEnabled == 1)	{
			NodesChanged++;
			//cout << "Calling SMALTAAnnounce\n";
			//cout << Prefix << " " << tmask << " " << nexthop << endl;
			SMALTA->SMALTAUpdate (Prefix, tmask, OldNH, 1);
		}
		else if (ZhangAggEnabled == 1)	{
			ZhangAgg->Update (Prefix, tmask, 1);
		}
	}
	else if (ret == 3)	{
		if (SMALTAEnabled == 1)	{
			NodesChanged++;
			//cout << "Calling SMALTAAnnounce\n";
			//cout << Prefix << " " << tmask << " " << nexthop << endl;
			SMALTA->SMALTAUpdate (Prefix,tmask, OldNH, 2);
		}
		else if (ZhangAggEnabled == 1)	{
			ZhangAgg->Update (Prefix, tmask, 2);
		}
	}
	else if (ret == 4)	{
		if (SMALTAEnabled == 1)	{
			NodesChanged++;
			//cout << "Calling SMALTAAnnounce\n";
			//cout << Prefix << " " << tmask << " " << nexthop << endl;
			SMALTA->SMALTAUpdate (Prefix,tmask, OldNH, 3);
		}
		else if (ZhangAggEnabled == 1)	{
			ZhangAgg->Update (Prefix, tmask, 2);
		}
	}
	//std::cout << (*next).Data.Routes.BestNH << " " << (*next).Data.Routes.BestASLen << "\n";
	return 0;
		
}

/*****************************************************************************************************
Function to Handle a Withdraw to the RIB. This may result in one of three cases:
a) There were no alternate routes after the withdraw.
b) There were alternate routes after the withdraw
c) The withdraw was for a node with non-original nexthop

returns 0 always

Parameters: 
Prefix:	Preix in 32 bit format 
mask:	Length of the Prefix
peerIP:	the nexthop of the route to be withdrawn
*****************************************************************************************************/


int cRIB::RIBWithdraw (unsigned int prefix, int mask, unsigned int peerIP)	{
	
	//cout << "withdraw\n";
	cRIBNode *next = root;
	int nbit = 32;
	int ret = 0;
	unsigned int bit = 0;
	int tmask = mask;
	unsigned int OldNH;
	//if (e > 5495611)
		//cout << prefix <<"\n";
	//std::cout << prefix << " " << mask << " " << peerIP << "\n";
	while (mask > 0)	{
		nbit--;
		bit = (prefix >> nbit) % 2;
		if ((*next).child[bit] == NULL)		// prefix already withdrawn
			return 0;
		else
			next = (*next).child[bit];
		mask--;
	}
	OldNH = (*next).Data.Routes.BestNH;
	ret = (*next).UnSetRoutes (peerIP, SMALTAEnabled, ZhangAggEnabled);
	//cout << "out of WITHD\n";
	if (ret == 1)	{
		//cout << "deleting next";	
		delete next;
	}
	else if (ret == 2)	{
		if (SMALTAEnabled == 1)		{
			NodesChanged++;
			//cout << "Calling SMALTA withdraw for withdraw\n";
			//cout << prefix << " " << tmask << " " << peerIP << endl;
			SMALTA->SMALTAUpdate (prefix, tmask, OldNH, 0);
		}
		else if (ZhangAggEnabled == 1)	{
			ZhangAgg->Update (prefix, tmask, 0);
		}
	}
	else if (ret == 3)	{
		if (SMALTAEnabled == 1)	{
			NodesChanged++;
			//cout << "Calling SMALTA Announce for withdraw\n";
			//cout << prefix << " " << tmask << " " << peerIP << endl;
			SMALTA->SMALTAUpdate (prefix, tmask, OldNH, 1);
		}
		else if (ZhangAggEnabled == 1)	{
			ZhangAgg->Update (prefix, tmask, 1);
		}
	}
	return 0;
}

/***********************************************************************************************************************
Interface fucntion for providing updates to the RIB

returns: Number of Nodes whose Best NextHop has changed upto the current update.

Parameters:
Prefix:	32 bit Prefix 
mask:	Length of the prefix
nexthop: nexthop if an announce, peerIP if withdraw
ASLength: AS path length of the announced route. Meaningless for withdraw.
MessageType:	1 for announce, 0 for withdraw
************************************************************************************************************************/

unsigned int cRIB::RIBUpdate (unsigned int Prefix, int mask, unsigned int nexthop, int ASLength, int MessageType)	{
	//std::cout << Prefix << " " << mask << " " << nexthop << " " << ASLength << " " << MessageType << "\n";
	//e++;
	
		
	if (MessageType == 1)	{
		
		RIBAnnounce (Prefix, mask, nexthop, ASLength);
	}
	else if (MessageType == 0)	{
		
		RIBWithdraw (Prefix, mask, nexthop);
	}
	else
		return 1;
	//cout << "Returning from RIB Update\n";
	return NodesChanged;
}

/***************************************************************************************************************
copies all prefixes, best nexthop corresponding to nodes with Original flag = 1 onto a file.
The file name is provided by the argument to the function FileName
returns 0

****************************************************************************************************************/


int cRIB::GetOriginalTable (string* FileName)	
{
	
  ofstream original ((*FileName).c_str());
  if (!original)	
    {
      std::cout << "could not open file\n";
      exit(1);
    }
  cRIBNode *tree = root;
  int depth = 0;
  std::vector<int> flags (33,0);
  std::cout << "In original tree \n";
  unsigned int prefix = 0;
	
  while  (tree != NULL)	
    {
      switch (flags[depth])	
	{	
	case 0:			// first time visit
				
	  if ((*tree).Data.Routes.BestNH != NULL_NEXTHOP && (*tree).Data.SMALTAFlags.Original == 1)	
	    {
	      original << prefix << " " << depth << " "	<< (*tree).Data.Routes.BestNH 
		       << " " << (*tree).Data.Routes.BestASLen << endl ;// << " " << (*tree).Data.Routes.NHASLenPair.size()
	      /*
		for (int i = 0; i < (*tree).Data.Routes.NextHop.size(); i++)	
		original << prefix << " " << depth << " " << (*tree).Data.Routes.NextHop[i] << " " << (*tree).Data.Routes.ASLength[i] << "\n";*/
	    }
	  flags[depth] = 1;
	  if ((*tree).child[LEFT] == NULL)	
	    {
	      flags[depth] = 2;
	      if ((*tree).child[RIGHT] == NULL)	
		{
		  flags[depth] = 0;
		  tree = (*tree).parent;
		  prefix = prefix >> (32- depth + 1) << (32 - depth + 1);
		  depth --;
		}
	      else	
		{
		  tree = (*tree).child[RIGHT];
		  depth++;
		  prefix = prefix | (1 << (32-depth));
		}
	    }
	  else	
	    {
	      tree = (*tree).child[LEFT];
	      depth++;
	      prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	    }
	  break;
	case 1:
	  //std::cout << "case 1 \n";
	  flags[depth] = 2;
	  if ((*tree).child[RIGHT] == NULL)	
	    {
	      flags[depth] = 0;
	      tree = (*tree).parent;
	      prefix = prefix >> (32-depth+1) << (32 - depth + 1);
	      depth--;
	    }
	  else	
	    {
	      tree = (*tree).child[RIGHT];
	      depth++;
	      prefix = prefix | (1 << (32 - depth));
	    }
	  break;
	case 2:
	  //std::cout << "case 2\n";
	  flags[depth] = 0;
	  tree = (*tree).parent;
	  prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	  depth--;
	  break;
	}
    }
  original.close();
  return 0;
	
}

/**********************************************************************************************************

Copies all prefixes, Best NextHop with INFIB flag set to 1 onto a file. The file name is given by
argument file. Returns the number of INFIB entries.

**********************************************************************************************************/

int cRIB::GetAggregatedTable (std::string *file)	{
	std::cout << "Regurgiating aggregated table\n";
	cRIBNode *next = root;
	std::vector <int> flags (33,0);
	int depth = 0;
	std::ofstream fp ((*file).c_str());
	unsigned int prefix = 0;
	int iter = 0;
	while (next != NULL)	{
		switch (flags[depth])	{
			case 0:
				if ((*next).Data.RIBFlags.INFIB == 1)	{
					iter++;
					//std::cout << (*next).Data.RIBFlags.INFIB << "\n";
					fp << prefix << " " << depth << " " << (*next).Data.Routes.BestNH
					/*<< " " << (*next).Data.SMALTAFlags.DistanceToParent << " " << 
					(*next).Data.SMALTAFlags.Original*/<< endl;
				}
				flags[depth] = 1;
				if (next->child[LEFT] == NULL)	{
					flags [depth] = 2;
					if (next->child[RIGHT] == NULL)	{
						flags [depth] = 0;
						next = next->parent;
						prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
						depth--;
					}
					else	{
						next = next->child[RIGHT];
							depth++;		
						prefix = prefix | (1 << (32-depth));
					
					}
				}
				else {
					next = next->child[LEFT];
					depth++;
					prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
				}
				break;
			
			case 1:
				flags[depth] = 2;
				if (next->child[RIGHT] == NULL)	{
					flags [depth] = 0;
					next = next->parent;	
					prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
					depth--;
				}
				else	{
					next = next->child[RIGHT];
					depth++;		
					prefix = prefix | (1 << (32-depth));
				}
				
				break;
			
			case 2:
				flags [depth] = 0;
				next = next->parent;	
				prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
				depth--;
		}

	}
	//std::cout << iter << "\n";	
	//std::cout << intersections << "\n";
	return iter;
}

/************************************************************************************************************************
Dual Nexthop RIB and functions
************************************************************************************************************************/

cRIBdualNHNode::cRIBdualNHNode (cRIBdualNHNode *par, cRIBdualNHNode *lchild, cRIBdualNHNode *rchild)	{
	parent = par;
	child[LEFT] = lchild;
	child[RIGHT] = rchild;
	GenNH = NULL_NEXTHOP;
	BestOrigNH = NULL_NEXTHOP;
	BestASLen = 0;
	DistanceToParent = -1;
}

int cRIBdualNHNode::SetRoutes (unsigned int NH, int ASLen)
{
  if(ALL_UPDATES_POTENT)
    {
      // the following two lines should be executed for ATT only
      BestOrigNH = NULL_NEXTHOP;
      BestASLen = 1000; // something infinite!
    }

  unsigned int CurrBestNH = BestOrigNH;
  int CurrBestASLen = BestASLen;
  int index = -1;
  if (CurrBestNH == NULL_NEXTHOP)
    {
      BestOrigNH = NH;
      BestASLen = ASLen;
      return 2;
    }
  else if (CurrBestNH == NH)	
    {
      if (CurrBestASLen == ASLen)
	return 0; // nothing changed in our best
      else if (CurrBestASLen > ASLen)	
	{
	  // our existing best just got improved
	  BestASLen = ASLen;
	  return 0;
	}
      else if (CurrBestASLen < ASLen)	
	{
	  // length of the existing best has increased
	  // need to check if some other route has become better
	  CurrBestASLen = ASLen;
	  for (unsigned int i = 0; i < NHASLenPair.size(); i++)	
	    {
	      if (CurrBestASLen > NHASLenPair[i].ASLen)	
		{
		  index = i;
		  CurrBestASLen = NHASLenPair[i].ASLen;
		}
	    }
	  if (index != -1)	
	    {
	      BestOrigNH = NHASLenPair[index].NH;
	      BestASLen = NHASLenPair[index].ASLen;
	      NHASLenPair.erase (NHASLenPair.begin() + index);
	      sNHASLenPair pair;
	      pair.NH = NH;
	      pair.ASLen = ASLen;
	      NHASLenPair.push_back (pair);
	      return 1;
	    }
	  else	
	    {
	      BestASLen = ASLen;
	      return 0;
	    }
	}
    }
  else // the announced nexthop is differnet form thenon-null current best
    {
      for (unsigned int i = 0; i < NHASLenPair.size(); i++)
	{	// if a change in ASLen of already existing NH
	  if (NHASLenPair[i].NH == NH)
	    {
	      if (NHASLenPair[i].ASLen == ASLen)
		return 0;
	      else	
		{ 
		  NHASLenPair[i].ASLen = ASLen;
		  index = i;
		  break;
		}
	    }
	}
      
      if (index != -1) // the announced route was already in the list of routes
	{
	  if (NHASLenPair[index].ASLen < CurrBestASLen)	
	    {
	      sNHASLenPair pair;
	      pair.NH = CurrBestNH;
	      pair.ASLen = CurrBestASLen;
	      NHASLenPair.push_back(pair);
	      BestOrigNH = NHASLenPair[index].NH;
	      BestASLen = NHASLenPair[index].ASLen;
	      NHASLenPair.erase(NHASLenPair.begin() + index);
	      return 1;
	    }
	}
      else	
	{
	  if (ASLen < CurrBestASLen)	
	    {
	      sNHASLenPair pair;
	      pair.NH = CurrBestNH;
	      pair.ASLen = CurrBestASLen;
	      NHASLenPair.push_back(pair);
	      BestOrigNH = NH;
	      BestASLen = ASLen;
	      return 1;
	    }
	  else	
	    {
	      sNHASLenPair pair;
	      pair.NH = NH;
	      pair.ASLen = ASLen;
	      NHASLenPair.push_back(pair);
	      return 0;
	    }
	}
    }
  
  return 0; // no return expected
}

int cRIBdualNHNode::UnSetRoutes (unsigned int peerIP, int ORTCUpdateEnabled)	{
  unsigned int CurrBestNH = BestOrigNH;
  int CurrBestASLen = BestASLen;
  int index = -1;
	
  if (CurrBestNH == NULL_NEXTHOP)
    return 0;
  else	
    {
      if(ALL_UPDATES_POTENT || (peerIP == CurrBestNH))
	{
	  // the withdraw is for the current best guy
	  if (NHASLenPair.size() == 0)	
	    {
	      BestOrigNH = NULL_NEXTHOP;
	      BestASLen = 0;
	      
	      if (child[LEFT] == NULL && child[RIGHT] == NULL)	
		{
		  if (ORTCUpdateEnabled == 1)
		    return 2; // let ORTC decide if this needs to be deleted
		  else
		    {
		      if (parent->child[LEFT] == this)
			parent->child[LEFT] = NULL;
		      else if (parent->child[RIGHT] == this)
			parent->child[RIGHT] = NULL;
		      else	
			{
			  cout << "error, cannot delete leaf after withdraw\n";
			  exit(1);
			}
		      return 1; // indicate to delete this node anyways 
		    }
		}
	      return 2;
	    }
	  
	  // find a new best path from the list of available paths
	  // start by assuming that the first one is the best
	  index = 0;	
	  CurrBestASLen = NHASLenPair[0].ASLen;
	  
	  for (unsigned int i = 0; i < NHASLenPair.size(); i++)	
	    {
	      if (NHASLenPair[i].ASLen < CurrBestASLen)	
		{
		  CurrBestASLen = NHASLenPair[i].ASLen;
		  index = i;
		}
	    }
	  
	  BestOrigNH = NHASLenPair[index].NH;
	  BestASLen = NHASLenPair[index].ASLen;
	  NHASLenPair.erase(NHASLenPair.begin()+index);
	  return 3;
	}
      else	
	{
	  for (unsigned int i = 0; i < NHASLenPair.size(); i++)	
	    {
	      if (NHASLenPair[i].NH == peerIP)	
		{
		  NHASLenPair.erase(NHASLenPair.begin() + i);
		  break;
		}
	    }
	  return 0;
	}			
    }
}

int cRIBdualNH::Update (unsigned int Prefix, int mask, unsigned int NextHop, int ASLength, int Message)	
{
  // debugging code only
  /*  if(mask == 21)
    {
      int findLoc = -1;

      for(unsigned int i = 0; i < debugVector.size(); i++)
	{
	  if(NextHop == debugVector[i])
	    findLoc = i;
	}
      
      if(Message == 1) // ANNOUNCE
	{
	  if(findLoc == -1)
	    {
	      debugVector.push_back(NextHop);
	    }
	  else
	    {
	      // it is already there
	    }
	}
      else // Message == 0 -- WITHDRAW
	{
	  if(findLoc == -1)
	    {
	      // never found --
	      cout << "Withdraw of something which never was announced" << endl;
	    }
	  else
	    {
	      debugVector.erase(debugVector.begin()+findLoc);
	    }
	}
    }
  cout << "(RIBUpdate) Vector size is: " << debugVector.size() << endl;
  */  
  // debugging finished
  

	if (Message == 0)	
		RIBWithdraw (Prefix, mask, NextHop);
	else if (Message == 1)
		RIBAnnounce (Prefix, mask, NextHop, ASLength);
	return 0; // no return expected
}

int cRIBdualNH::RIBWithdraw (unsigned int prefix, int mask, unsigned int peerIP)	{
	cRIBdualNHNode *next = root;
	int nbit = 32;
	int ret = 0;
	unsigned int bit = 0;
	int tmask = mask;
	unsigned int OldNH;
	while (mask > 0)	{
		nbit--;
		bit = (prefix >> nbit) % 2;
		if ((*next).child[bit] == NULL)		// prefix already withdrawn
			return 0;
		else
			next = (*next).child[bit];
		mask--;
	}
	OldNH = (*next).BestOrigNH;
	ret = (*next).UnSetRoutes (peerIP, ORTCUpdateEnabled);
		if (ret == 1)	{
		delete next;
	}
	else if (ret == 2)	{
		if (ORTCUpdateEnabled == 1)		{
			ORTCUpdate->Update (prefix, tmask, OldNH, 0);
		}
		
	}
	else if (ret == 3)	{
		if (ORTCUpdateEnabled == 1)	{
			ORTCUpdate->Update (prefix, tmask, OldNH, 1);
		}
	}
	return 0;
}

int cRIBdualNH::RIBAnnounce (unsigned int Prefix, int mask, unsigned int NextHop, int ASLength)	{
	cRIBdualNHNode *next = root;
	int nbit = 32;
	int ret;
	int tmask = mask;
	unsigned int OldNH;
	unsigned int bit = 0;
	
	while (mask > 0)	{
		nbit--;
		bit = (Prefix >> nbit) % 2;
		if ((*next).child[bit] == NULL)	{ 
			cRIBdualNHNode* NewNode = new cRIBdualNHNode (next, NULL, NULL);
			(*next).child[bit] = NewNode;
			next = NewNode;
		}
		else	{
			next = (*next).child[bit];
		}
		mask--;
	}
	OldNH = (*next).BestOrigNH;

	ret = (*next).SetRoutes (NextHop, ASLength);
 	if (ret == 1)	{
		if (ORTCUpdateEnabled == 1)	{
			ORTCUpdate->Update (Prefix, tmask, OldNH, 1);
		}
	}
	
	else if (ret == 2)	{
		if (ORTCUpdateEnabled == 1)	{
			ORTCUpdate->Update (Prefix,tmask, OldNH, 2);
		}
		
	}
	
	return 0;
		
}

int cRIBdualNH::GetAggregatedTable (std::string *file)	
{
  cRIBdualNHNode *next = root;
  vector <int> flags (33,0);
  int depth = 0;
  bool writeToFile = (file != NULL);
  ofstream fp;

  if(writeToFile)
    fp.open((*file).c_str());

  unsigned int prefix = 0;
  int iter = 0;
  while (next != NULL)	
    {
      switch (flags[depth])	
	{
	case 0:
	  if (next->GenNH != NULL_NEXTHOP)	
	    {
	      iter++;
	      if(writeToFile)
		fp << prefix << " " << depth << " " << next->GenNH << endl;
	    }
	  flags[depth] = 1;
	  if (next->child[LEFT] == NULL)	
	    {
	      flags [depth] = 2;
	      if (next->child[RIGHT] == NULL)	
		{
		  flags [depth] = 0;
		  next = next->parent;
		  prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
		  depth--;
		}
	      else	
		{
		  next = next->child[RIGHT];
		  depth++;		
		  prefix = prefix | (1 << (32-depth));
					
		}
	    }
	  else 
	    {
	      next = next->child[LEFT];
	      depth++;
	      prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	    }
	  break;
			
	case 1:
	  flags[depth] = 2;
	  if (next->child[RIGHT] == NULL)	
	    {
	      flags [depth] = 0;
	      next = next->parent;	
	      prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	      depth--;
	    }
	  else	
	    {
	      next = next->child[RIGHT];
	      depth++;		
	      prefix = prefix | (1 << (32-depth));
	    }
				
	  break;
			
	case 2:
	  flags [depth] = 0;
	  next = next->parent;	
	  prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	  depth--;
	}

    }
	
  return iter;
}

int cRIBdualNH::GetOriginalTable (std::string *file)
{
  cRIBdualNHNode *next = root;
  vector <int> flags (33,0);
  int depth = 0;
  bool writeToFile = (file != NULL);
  ofstream fp;

  if(writeToFile)
    fp.open((*file).c_str());
  
  unsigned int prefix = 0;
  int iter = 0;
  while (next != NULL)	
    {
      switch (flags[depth])	
	{
	case 0:
	  if (next->BestOrigNH != NULL_NEXTHOP)	
	    {
	      iter++;
	      if(writeToFile)
		fp << prefix << " " << depth << " " 
		   << next->BestOrigNH << " " << next->BestASLen << endl;
	    }
	  flags[depth] = 1;
	  if (next->child[LEFT] == NULL)	
	    {
	      flags [depth] = 2;
	      if (next->child[RIGHT] == NULL)	
		{
		  flags [depth] = 0;
		  next = next->parent;
		  prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
		  depth--;
		}
	      else	
		{
		  next = next->child[RIGHT];
		  depth++;		
		  prefix = prefix | (1 << (32-depth));
					
		}
	    }
	  else 
	    {
	      next = next->child[LEFT];
	      depth++;
	      prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	    }
	  break;
			
	case 1:
	  flags[depth] = 2;
	  if (next->child[RIGHT] == NULL)	
	    {
	      flags [depth] = 0;
	      next = next->parent;	
	      prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	      depth--;
	    }
	  else	
	    {
	      next = next->child[RIGHT];
	      depth++;		
	      prefix = prefix | (1 << (32-depth));
	    }
				
	  break;
			
	case 2:
	  flags [depth] = 0;
	  next = next->parent;	
	  prefix = prefix >> (32 - depth + 1) << (32 - depth + 1);
	  depth--;
	}
    }
  fp.close();
  cout << iter << endl;
  return iter; // no return value expected
}
